@@ -8,6 +8,7 @@ use IO::String;
 use CGI::Info;
 use Carp;
 use HTTP::Date;
+use Text::Diff;	# For debugging
 
 =head1 NAME
 
@@ -15,11 +16,11 @@ CGI::Buffer - Verify, Cache and Optimise CGI Output
 
 =head1 VERSION
 
-Version 0.75
+Version 0.76
 
 =cut
 
-our $VERSION = '0.75';
+our $VERSION = '0.76';
 
 =head1 SYNOPSIS
 
@@ -56,6 +57,9 @@ But that's simple:
 
     # ...
 
+To temporarily prevent the use of server-side caches, for example whilst
+debugging before publishing a code change, set the NO_CACHE environment variable
+to any non-zero value.
 If you get errors about Wide characters in print it means that you've
 forgotten to emit pure HTML on non-ascii characters.
 See L<HTML::Entities>.
@@ -84,7 +88,7 @@ our $status;
 our $script_mtime;
 our $cobject;
 our($x_cache, $buf, $headers, $header, $body, @content_type, $etag,
-	$send_body, @o);
+	$send_body, @o, $encode_loaded);
 
 BEGIN {
 	# use Exporter();
@@ -127,7 +131,9 @@ END {
 	($headers, $body) = split /\r?\n\r?\n/, $buf, 2;
 
 	unless($headers || is_cached()) {
-		# There was no output
+		if($logger) {
+			$logger->debug('There was no output');
+		}
 		return;
 	}
 	if($ENV{'REQUEST_METHOD'} && ($ENV{'REQUEST_METHOD'} eq 'HEAD')) {
@@ -137,13 +143,7 @@ END {
 	}
 
 	if($headers) {
-		foreach my $header (split(/\r?\n/, $headers)) {
-			my ($header_name, $header_value) = split /\:\s*/, $header, 2;
-			if (lc($header_name) eq 'content-type') {
-				@content_type = split /\//, $header_value, 2;
-				last;
-			}
-		}
+		_set_content_type($headers);
 	}
 
 	if(defined($body) && ($body eq '')) {
@@ -159,14 +159,22 @@ END {
 			# require HTML::Clean;
 			require HTML::Packer;	# Overkill using HTML::Clean and HTML::Packer...
 
+			if($logger) {
+				$logger->trace('Packer');
+			}
+
 			my $oldlength = length($body);
 			my $newlength;
 
-			while(1) {
-				$body = _optimise_content($body);
-				$newlength = length($body);
-				last if ($newlength >= $oldlength);
-				$oldlength = $newlength;
+			if($optimise_content == 1) {
+				_optimise_content();
+			} else {
+				while(1) {
+					_optimise_content();
+					$newlength = length($body);
+					last if ($newlength >= $oldlength);
+					$oldlength = $newlength;
+				}
 			}
 
 			# If we're on http://www.example.com and have a link
@@ -235,6 +243,9 @@ END {
 			require HTML::Lint;
 			HTML::Lint->import;
 
+			if($logger) {
+				$logger->trace('Lint');
+			}
 			my $lint = HTML::Lint->new();
 			$lint->parse($body);
 
@@ -264,7 +275,6 @@ END {
 	# Generate the eTag before compressing, since the compressed data
 	# includes the mtime field which changes thus causing a different
 	# Etag to be generated
-	my $encode_loaded;
 	if($ENV{'SERVER_PROTOCOL'} &&
 	  ($ENV{'SERVER_PROTOCOL'} eq 'HTTP/1.1') &&
 	  $generate_etag && defined($body)) {
@@ -283,6 +293,8 @@ END {
 				if($logger) {
 					$logger->debug('Set status to 304');
 				}
+			} elsif($logger) {
+				$logger->debug(diff(\$body, \$cache->get(generate_key())));
 			}
 		}
 	}
@@ -290,7 +302,7 @@ END {
 	my $encoding = _should_gzip();
 	my $unzipped_body = $body;
 
-	if((length($encoding) > 0) && defined($body)) {
+	if(defined($unzipped_body)) {
 		my $range = $ENV{'Range'} ? $ENV{'Range'} : $ENV{'HTTP_RANGE'};
 
 		if($range && !$cache) {
@@ -306,22 +318,7 @@ END {
 				$unzipped_body = $body;
 			}
 		}
-		if(length($body) >= MIN_GZIP_LEN) {
-			require Compress::Zlib;
-			Compress::Zlib->import;
-
-			# Avoid 'Wide character in memGzip'
-			unless($encode_loaded) {
-				require Encode;
-				$encode_loaded = 1;
-			}
-			my $nbody = Compress::Zlib::memGzip(\Encode::encode_utf8($body));
-			if(length($nbody) < length($body)) {
-				$body = $nbody;
-				push @o, "Content-Encoding: $encoding";
-				push @o, "Vary: Accept-Encoding";
-			}
-		}
+		_compress({ encoding => $encoding });
 	}
 
 	if($cache) {
@@ -337,6 +334,7 @@ END {
 				if(defined($cobject)) {
 					$cache_hash = Storable::thaw($cobject->value());
 					$headers = $cache_hash->{'headers'};
+					_set_content_type($headers);
 					@o = ("X-CGI-Buffer-$VERSION: Hit");
 					if($info) {
 						my $host_name = $info->host_name();
@@ -407,26 +405,14 @@ END {
 						}
 					}
 				}
-				if($send_body && (length($encoding) > 0)) {
-					if(length($body) >= MIN_GZIP_LEN) {
-						require Compress::Zlib;
-						Compress::Zlib->import;
-
-						# Avoid 'Wide character in memGzip'
-						unless($encode_loaded) {
-							require Encode;
-							$encode_loaded = 1;
-						}
-						if($generate_etag && !defined($etag) && ((!defined($headers)) || ($headers !~ /^ETag: /m))) {
-							$etag = '"' . Digest::MD5->new->add(Encode::encode_utf8($body))->hexdigest() . '"';
-						}
-						my $nbody = Compress::Zlib::memGzip(\Encode::encode_utf8($body));
-						if(length($nbody) < length($body)) {
-							$body = $nbody;
-							push @o, "Content-Encoding: $encoding";
-							push @o, "Vary: Accept-Encoding";
-						}
+			}
+			if($status == 200) {
+				$encoding = _should_gzip();
+				if($send_body) {
+					if($generate_etag && !defined($etag) && ((!defined($headers)) || ($headers !~ /^ETag: /m))) {
+						$etag = '"' . Digest::MD5->new->add(Encode::encode_utf8($body))->hexdigest() . '"';
 					}
+					_compress({ encoding => $encoding });
 				}
 			}
 			my $cannot_304 = !$generate_304;
@@ -438,6 +424,9 @@ END {
 				}
 			}
 			if($ENV{'HTTP_IF_NONE_MATCH'} && $send_body && ($status != 304) && $generate_304) {
+				if($logger) {
+					$logger->debug("Compare $ENV{HTTP_IF_NONE_MATCH} with $etag");
+				}
 				if(defined($etag) && ($etag eq $ENV{'HTTP_IF_NONE_MATCH'}) && ($status == 200)) {
 					push @o, "Status: 304 Not Modified";
 					$send_body = 0;
@@ -458,12 +447,13 @@ END {
 				}
 				if(($status == 200) && $generate_last_modified) {
 					if($logger) {
-						$logger->debug('Set Last-Modified to ' . HTTP::Date::time2str($cobject->created_at()));
+						$logger->debug('Set Last-Modified to ', HTTP::Date::time2str($cobject->created_at()));
 					}
 					push @o, "Last-Modified: " . HTTP::Date::time2str($cobject->created_at());
 				}
 			}
 		} else {
+			# Not in the server side cache
 			if($status == 200) {
 				unless($cache_age) {
 					# It would be great if CHI::set()
@@ -472,7 +462,7 @@ END {
 					$cache_age = '10 minutes';
 				}
 				$cache_hash->{'body'} = $unzipped_body;
-				if(@o && scalar(@o)) {
+				if(defined(@o) && scalar(@o)) {
 					# Remember, we're storing the UNzipped
 					# version in the cache
 					my $c;
@@ -498,10 +488,13 @@ END {
 				if($generate_etag && defined($etag)) {
 					$cache_hash->{'etag'} = $etag
 				}
-				$cache->set($key, Storable::freeze($cache_hash), $cache_age);
+				# TODO: Support the Expires header
+				# if($headers !~ /^Expires: /m))) {
+				# }
 				if($logger) {
-					$logger->debug("store $key in the cache");
+					$logger->debug("Store $key in the cache, age = $cache_age ", length($cache_hash->{'body'}), ' bytes');
 				}
+				$cache->set($key, Storable::freeze($cache_hash), $cache_age);
 				if($generate_last_modified) {
 					$cobject = $cache->get_object($key);
 					if(defined($cobject)) {
@@ -648,10 +641,16 @@ sub _check_modified_since {
 		return;
 	}
 	if($age > $s) {
+		if($logger) {
+			$logger->debug('_check_modified_since: script has been modified');
+		}
 		# Script has been updated so it may produce different output
 		return;
 	}
 
+	if($logger) {
+		$logger->debug("_check_modified_since: Compare $$params{modified} with $s");
+	}
 	if($$params{modified} <= $s) {
 		push @o, "Status: 304 Not Modified";
 		$status = 304;
@@ -663,15 +662,14 @@ sub _check_modified_since {
 }
 
 sub _optimise_content {
-	my $body = shift;
-
+	# FIXME: regex bad, HTML parser good
 	# Regexp::List - wow!
-	$body =~ s/(\s+|\r)\n|\n\+/\n/gs;
+	$body =~ s/(((\s+|\r)\n|\n(\s+|\+)))/\n/g;
 	# $body =~ s/\r\n/\n/gs;
 	# $body =~ s/\s+\n/\n/gs;
 	# $body =~ s/\n+/\n/gs;
-	$body =~ s/\<\/option\>\s\<option/\<\/option\>\<option/gis;
-	$body =~ s/\<\/div\>\s\<div/\<\/div\>\<div/gis;
+	# $body =~ s/\n\s+|\s+\n/\n/g;
+	$body =~ s/\<\/div\>\s+\<div/\<\/div\>\<div/gis;
 	# $body =~ s/\<\/p\>\s\<\/div/\<\/p\>\<\/div/gis;
 	# $body =~ s/\<div\>\s+/\<div\>/gis;	# Remove spaces after <div>
 	$body =~ s/(<div>\s+|\s+<div>)/<div>/gis;
@@ -684,8 +682,6 @@ sub _optimise_content {
 	$body =~ s/<body>\s+/<body>/is;
 	$body =~ s/\s+\<\/html/\<\/html/is;
 	$body =~ s/\s+\<\/body/\<\/body/is;
-	$body =~ s/\n\s+|\s+\n/\n/g;
-	$body =~ s/\t+/ /g;
 	$body =~ s/\s(\<.+?\>\s\<.+?\>)/$1/;
 	# $body =~ s/(\<.+?\>\s\<.+?\>)\s/$1/g;
 	$body =~ s/\<p\>\s/\<p\>/gi;
@@ -697,22 +693,23 @@ sub _optimise_content {
 	$body =~ s/\<br\s?\/?\>\s?\<p\>/\<p\>/gi;
 	$body =~ s/\<br\>\s/\<br\>/gi;
 	$body =~ s/\<br\s?\/\>\s/\<br \/\>/gi;
-	$body =~ s/ +/ /gs;	# Remove duplicate space, don't use \s+ it breaks JavaScript
+	$body =~ s/[ \t]+/ /gs;	# Remove duplicate space, don't use \s+ it breaks JavaScript
 	$body =~ s/\s\<p\>/\<p\>/gi;
 	$body =~ s/\s\<script/\<script/gi;
 	$body =~ s/(<script>\s|\s<script>)/<script>/gis;
 	$body =~ s/(<\/script>\s|\s<\/script>)/<\/script>/gis;
 	$body =~ s/\<td\>\s/\<td\>/gi;
-	$body =~ s/\s?\<a\shref="(.+?)"\>\s?/ <a href="$1">/gis;
+	$body =~ s/\s+\<a\shref="(.+?)"\>\s?/ <a href="$1">/gis;
 	$body =~ s/\s?<a\shref=\s"(.+?)"\>/ <a href="$1">/gis;
-	$body =~ s/(\s?<hr>\s|\s<hr>\s?)/<hr>/gis;
+	$body =~ s/\s+<\/a\>\s+/<\/a> /gis;
+	$body =~ s/(\s?<hr>\s+|\s+<hr>\s?)/<hr>/gis;
 	# $body =~ s/\s<hr>/<hr>/gis;
 	# $body =~ s/<hr>\s/<hr>/gis;
-	$body =~ s/<\/li>\s<li>/<\/li><li>/gis;
-	$body =~ s/<\/li>\s<\/ul>/<\/li><\/ul>/gis;
-	$body =~ s/<ul>\s<li>/<ul><li>/gis;
-
-	return $body;
+	$body =~ s/<\/li>\s+<li>/<\/li><li>/gis;
+	$body =~ s/<\/li>\s+<\/ul>/<\/li><\/ul>/gis;
+	$body =~ s/<ul>\s+<li>/<ul><li>/gis;
+	$body =~ s/\<\/option\>\s+\<option/\<\/option\>\<option/gis;
+	$body =~ s/<title>\s*(.+?)\s*<\/title>/<title>$1<\/title>/is;
 }
 
 # Create a key for the cache
@@ -729,9 +726,40 @@ sub _generate_key {
 	my $key = $info->browser_type() . '::' . $info->domain_name() . '::' . $info->script_name() . '::' . $info->as_string();
 	if($ENV{'HTTP_COOKIE'}) {
 		# Different states of the client are stored in different caches
-		$key .= '::' . $ENV{'HTTP_COOKIE'};
+		# Don't put different Google Analytics in different caches, and anyway they
+		# would be wrong
+		foreach my $cookie(split(/;/, $ENV{'HTTP_COOKIE'})) {
+			unless($cookie =~ /^__utm[abcz]/) {
+				$key .= "::$cookie";
+			}
+		}
+	}
+
+	# Honour the Vary headers
+	if($headers && ($headers =~ /^Vary: .*$/m)) {
+		if(defined($logger)) {
+			$logger->debug('Found Vary header');
+		}
+		foreach my $h1(split(/\r?\n/, $headers)) {
+			my ($h1_name, $h1_value) = split /\:\s*/, $h1, 2;
+			if(lc($h1_name) eq 'vary') {
+				foreach my $h2(split(/\r?\n/, $headers)) {
+					my ($h2_name, $h2_value) = split /\:\s*/, $h2, 2;
+					if($h2_name eq $h1_value) {
+						$key .= '::' . $h2_value;
+						last;
+					}
+				}
+			}
+		}
 	}
 	$key =~ s/\//::/g;
+	$key =~ s/::::/::/g;
+	$key =~ s/::$//;
+	if(defined($logger)) {
+		$logger->trace("Returning $key");
+	}
+	$cache_key = $key;
 	return $key;
 }
 
@@ -843,7 +871,19 @@ sub init {
 			}
 		}
 		$cache_age ||= $params{cache_age};
-		$cache = $params{cache};
+
+		if((!defined($params{cache})) && defined($cache)) {
+			if(defined($logger)) {
+				if($cache_key) {
+					$logger->debug("disabling cache $cache_key");
+				} else {
+					$logger->debug('disabling cache');
+				}
+			}
+			$cache = undef;
+		} else {
+			$cache = $params{cache};
+		}
 		if(defined($params{cache_key})) {
 			$cache_key = $params{cache_key};
 		}
@@ -854,7 +894,7 @@ sub import {
 	# my $class = shift;
 	shift;
 
-	return unless @_;
+	return unless defined(@_);
 
 	init(@_);
 }
@@ -891,6 +931,7 @@ sub can_cache {
 		if(defined($logger)) {
 			$logger->debug("cache_control = $control");
 		}
+		# TODO: check Authorization header not present
 		if(($control eq 'no-store') ||
 		       ($control eq 'no-cache') ||
 		       ($control eq 'max-age=0') ||
@@ -1026,7 +1067,7 @@ sub _should_gzip {
 		my $accept = lc($ENV{'HTTP_ACCEPT_ENCODING'} ? $ENV{'HTTP_ACCEPT_ENCODING'} : $ENV{'HTTP_TE'});
 		foreach my $encoding ('x-gzip', 'gzip') {
 			$_ = $accept;
-			if(@content_type && $content_type[0]) {
+			if(defined(@content_type) && $content_type[0]) {
 				if (m/$encoding/i && (lc($content_type[0]) eq 'text')) {
 					return $encoding;
 				}
@@ -1041,6 +1082,44 @@ sub _should_gzip {
 	return '';
 }
 
+sub _set_content_type
+{
+	my $headers = shift;
+
+	foreach my $header (split(/\r?\n/, $headers)) {
+		my ($header_name, $header_value) = split /\:\s*/, $header, 2;
+		if (lc($header_name) eq 'content-type') {
+			@content_type = split /\//, $header_value, 2;
+			last;
+		}
+	}
+}
+
+sub _compress {
+	my %params = (ref($_[0]) eq 'HASH') ? %{$_[0]} : @_;
+
+	my $encoding = $params{encoding};
+
+	if((length($encoding) == 0) || (length($body) < MIN_GZIP_LEN)) {
+		return;
+	}
+
+	require Compress::Zlib;
+	Compress::Zlib->import;
+
+	# Avoid 'Wide character in memGzip'
+	unless($encode_loaded) {
+		require Encode;
+		$encode_loaded = 1;
+	}
+	my $nbody = Compress::Zlib::memGzip(\Encode::encode_utf8($body));
+	if(length($nbody) < length($body)) {
+		$body = $nbody;
+		push @o, "Content-Encoding: $encoding";
+		push @o, "Vary: Accept-Encoding";
+	}
+}
+
 =head1 AUTHOR
 
 Nigel Horne, C<< <njh at bandsman.co.uk> >>
